import argparse
import os
import sys
from typing import Tuple

from docker_ascii_map import __version__


# This code is taken from Django
def supports_color():
    """
    Returns True if the running system's terminal supports color, and False
    otherwise.
    """
    plat = sys.platform
    supported_platform = plat != 'Pocket PC' and (plat != 'win32' or 'ANSICON' in os.environ)
    # isatty is not always implemented, #6223.
    is_a_tty = hasattr(sys.stdout, 'isatty') and sys.stdout.isatty()
    if not supported_platform or not is_a_tty:
        return False
    return True


def get_input_parameters() -> Tuple[bool]:
    parser = argparse.ArgumentParser(description='Display the docker host contents on a visual map.')
    parser.add_argument('-v', '--version', action='version', version='%(prog)s ' + __version__)
    parser.add_argument('-c', '--color', action='store_const', const='color', help='Use color in map display')
    parser.add_argument('-m', '--mono', action='store_const', const='mono', help='Render the map in monochrome')
    parser.add_argument('-f', '--follow', action='store_const', const='follow',
                        help='Follow status changes continuously')

    args = parser.parse_args()
    color_mode = supports_color()

    if args.color:
        color_mode = True

    if args.mono:
        color_mode = False

    return color_mode, args.follow is not None



from typing import List

import docker


class PortMapping:
    def __init__(self, private_port: int, public_port: int):
        self.private_port = private_port
        self.public_port = public_port

    def __repr__(self):
        return '%d:%d' % (self.public_port, self.private_port)


class VolumeMapping:
    def __init__(self, rw: bool, source: str, destination: str):
        self.rw = rw
        self.source = source
        self.destination = destination

    def __repr__(self):
        return '%s:%s' % (self.source, self.destination)


class Container:
    def __init__(self, name: str, status: str, networks: List[str], image: str,
                 ports: List[PortMapping] = None,
                 volumes: List[VolumeMapping] = None):
        self.name = name
        self.status = status
        self.networks = networks
        self.image = image
        self.ports = ports if ports is not None else []
        self.volumes = volumes if volumes is not None else []

    def is_running(self):
        return self.status == 'running'

    def __repr__(self) -> str:
        return '%s - %s - %s %s %s' % (self.name, self.status, self.networks, self.ports, self.volumes)


class Configuration:
    def __init__(self, containers: List[Container]):
        self.containers = containers

    def __repr__(self):
        return str([repr(c) for c in self.containers])


class ConfigParser:
    def __init__(self):
        self._client = docker.from_env(assert_hostname=False)

    def get_configuration(self) -> Configuration:
        containers = []

        for cinfo in self._client.containers(all=True):
            name = cinfo['Names'][0]

            if name[0] == '/':
                name = name[1:]

            status = cinfo['State']
            image = cinfo['Image']
            networks = [n for n in cinfo['NetworkSettings']['Networks'].keys()]
            networks.sort()
            ports = [PortMapping(p['PrivatePort'], p['PublicPort'])
                     for p in cinfo['Ports'] if 'PublicPort' in p.keys()]
            volumes = [VolumeMapping(rw=v['RW'], source=v['Source'], destination=v['Destination'])
                       for v in cinfo['Mounts']]
            containers.append(Container(name, status, networks, image, ports, volumes))

        return Configuration(containers)



from typing import List, Tuple

from docker_ascii_map.raster import Raster


class Size:
    def __init__(self, width: int, height: int):
        self.width = width
        self.height = height

    def __repr__(self):
        return 'Size{w:%d,h:%d}' % (self.width, self.height)

    def __eq__(self, other):
        return self.__dict__ == other.__dict__


class Hints:
    def __init__(self, size: Size):
        self.size = size


class Widget:
    def render(self, hints: Hints = None) -> Raster:
        pass

    def preferred_size(self) -> Size:
        pass


class Padding(Widget):
    def __init__(self, component: Widget, amount_tl: Size, amount_br: Size = None):
        self._component = component
        self._amount_tl = amount_tl
        self._amount_br = amount_br if amount_br is not None else amount_tl

    def preferred_size(self) -> Size:
        nested = self._component.preferred_size()
        return Size(nested.width + self._amount_tl.width + self._amount_br.width,
                    nested.height + self._amount_tl.height + self._amount_br.height)

    def render(self, hints: Hints = None) -> Raster:
        r = Raster()
        wrapper_size = self.preferred_size()

        for y in range(wrapper_size.height):
            for x in range(wrapper_size.width):
                r.write(x, y, ' ', self)

        cmp_raster = self._component.render()
        r.write(self._amount_tl.width, self._amount_tl.height, cmp_raster)

        return r


class Border(Widget):
    def __init__(self, component: Widget, title: str = ''):
        self._component = Padding(component, Size(1, 0))
        self._title = title

    def preferred_size(self) -> Size:
        nested = self._component.preferred_size()
        return Size(nested.width + 2, nested.height + 2)

    def render(self, hints: Hints = None) -> Raster:
        cmp_raster = self._component.render()
        cmp_size = self._component.preferred_size()
        cmp_w, cmp_h = cmp_size.width, cmp_size.height

        if hints is not None:
            cmp_w, cmp_h = hints.size.width - 2, hints.size.height - 2

        min_width = 2 + len(self._title)
        width = max(min_width, cmp_w)

        r = Raster()

        for y in range(cmp_h + 2):
            r.write(0, y, '|', origin=self, color='white', attrs=['dark'])
            r.write(width + 1, y, '|', origin=self, color='white', attrs=['dark'])

        for x in range(width + 2):
            r.write(x, 0, '-', origin=self, color='white', attrs=['dark'])
            r.write(x, cmp_h + 1, '-', origin=self, color='white', attrs=['dark'])

        r.write(0, 0, '+', origin=self, color='white', attrs=['dark'])
        r.write(width + 1, 0, '+', origin=self, color='white', attrs=['dark'])
        r.write(0, cmp_h + 1, '+', origin=self, color='white', attrs=['dark'])
        r.write(width + 1, cmp_h + 1, '+', origin=self, color='white', attrs=['dark'])

        if len(self._title) > 0:
            r.write(2, 0, ' ' + self._title + ' ', origin=self, color='white')

        r.write(1, 1, cmp_raster)

        return r


class VBox(Widget):
    def __init__(self, content: List[Widget]):
        self._content = content

    def preferred_size(self) -> Size:
        w, h = 0, 0

        for content in self._content:
            ps = content.preferred_size()
            w = max(ps.width, w)
            h += ps.height

        return Size(w, h)

    def render(self, hints: Hints = None):
        r = Raster()
        max_width = self.preferred_size().width

        for content in self._content:
            hints = Hints(Size(max_width, content.preferred_size().height))
            contentraster = content.render(hints)
            r.write(0, r.size()[1], contentraster)

        return r


class HBox(Widget):
    def __init__(self, content: List[Widget]):
        self._content = content

    def preferred_size(self) -> Size:
        w, h = 0, 0

        for content in self._content:
            ps = content.preferred_size()
            w += ps.width
            h = max(ps.height, h)

        return Size(w, h)

    def render(self, hints: Hints = None):
        r = Raster()

        for content in self._content:
            hints = Hints(content.preferred_size())
            contentraster = content.render(hints)
            r.write(r.size()[0], 0, contentraster)

        return r


class Paragraph(Widget):
    def __init__(self, lines: List[str], color: str = None, attrs: List[str] = None):
        self._lines = lines
        self._color = color
        self._attrs = attrs

    def preferred_size(self) -> Size:
        return Size(max([len(l) for l in self._lines]), len(self._lines))

    def render(self, hints: Hints = None):
        r = Raster()

        for l in self._lines:
            r.write(0, r.size()[1], l, origin=self, color=self._color, attrs=self._attrs)

        return r


class Links(Widget):
    def __init__(self, left: Widget, right: Widget,
                 links: List[Tuple[Widget, Widget]],
                 align: bool = False):
        self._left = left
        self._right = right
        self._links = links
        self._align = align

    def preferred_size(self) -> Size:
        return HBox([self._left, self._right]).preferred_size()

    def render(self, hints: Hints = None):
        dy_values = []
        dy_average = 0
        root = HBox([self._left, self._right])
        raster = root.render(hints)

        for w_src, w_dst in self._links:
            dst_x, dst_y, src_x, src_y = self._compute_link_coords(raster, w_dst, w_src)
            raster.draw_line(src_x, src_y, dst_x, dst_y)
            dy_values.append(dst_y - src_y)

        if self._align and len(dy_values) > 0:
            dy_average = int(sum(dy_values) / len(dy_values))

        if dy_average != 0:
            if dy_average < 0:
                root = HBox([self._left, Padding(self._right, Size(0, -dy_average))])
            else:
                root = HBox([Padding(self._left, Size(0, dy_average)), self._right])

            raster = root.render(hints)

            for w_src, w_dst in self._links:
                dst_x, dst_y, src_x, src_y = self._compute_link_coords(raster, w_dst, w_src)
                raster.draw_line(src_x, src_y, dst_x, dst_y)
                dy_values.append(dst_y - src_y)

        return raster

    def _compute_link_coords(self, raster, w_dst, w_src):
        bounds_src = raster.origin_bounds(w_src)
        bounds_dst = raster.origin_bounds(w_dst)
        src_x = bounds_src.x + bounds_src.w
        src_y = int(bounds_src.y + bounds_src.h / 2) - 1
        dst_x = bounds_dst.x
        dst_y = int(bounds_dst.y + bounds_dst.h / 2) - 1
        return dst_x, dst_y, src_x, src_y


class Annotations(Widget):
    def __init__(self, content: Widget, annotations: List[Tuple[Widget, str, str, List[str]]]):
        self._content = Padding(content, Size(3, 0), Size(0, 0)) if len(annotations) > 0 else content
        self._annotations = annotations
        self._width = max([0] + [len(a[1]) for a in annotations])

    def preferred_size(self) -> Size:
        content_size = self._content.preferred_size()
        return Size(content_size.width + self._width, content_size.height)

    def render(self, hints: Hints = None):
        raster = Raster()
        raster.write(self._width, 0, self._content.render(hints))

        widget_annotations_map = {}

        used_y = []

        for widget, annotation_text, color, attrs in self._annotations:
            if widget not in widget_annotations_map.keys():
                widget_annotations_map[widget] = []

            widget_annotations_map[widget].append((annotation_text, color, attrs))

        for widget, annotations in widget_annotations_map.items():
            bounds = raster.origin_bounds(widget)
            y = int(bounds.y + (bounds.h - len(annotations)) / 2)

            for annotation_text, color, attrs in annotations:
                while y in used_y:
                    y += 1

                used_y.append(y)

                raster.write(0, y, annotation_text, color=color, attrs=attrs)
                raster.write(self._width, y, ' ]-', color='white')
                raster.draw_line(self._width + 3, y, bounds.x - 1, y, color='white')

        return raster



from typing import Tuple, List

import termcolor


class Boundary:
    def __init__(self, x: int, y: int, w: int, h: int):
        self.x = x
        self.y = y
        self.w = w
        self.h = h

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def __repr__(self):
        return '%d,%d %dx%d' % (self.x, self.y, self.w, self.h)


class RasterCell:
    def __init__(self, character: str = ' ', origin: object = None, color: str = None, attrs: List[str] = None):
        self.character = character
        self.origin = origin
        self.color = color
        self.attrs = attrs

    def __eq__(self, other):
        return self.__dict__ == other.__dict__


class Raster:
    def __init__(self):
        self._cells = []
        self._default = RasterCell()

    def write(self, x: int, y: int, text, origin: object = None, color: str = None, attrs: List[str] = None):
        if type(text) is Raster:
            rastersize_x, rastersize_y = text.size()

            for ry in range(rastersize_y):
                for rx in range(rastersize_x):
                    cell = text.get(rx, ry)
                    self.write(x + rx, y + ry, origin=cell.origin, color=cell.color, attrs=cell.attrs,
                               text=cell.character)
        else:
            self._expand(x + len(text), y + 1)

            for i in range(len(text)):
                self._cells[y][x + i] = RasterCell(character=text[i], origin=origin, color=color, attrs=attrs)

    def draw_line(self, src_x, src_y, dst_x, dst_y, color: str = None, attrs: List[str] = None):
        med_x = int((src_x + dst_x) / 2)

        for x in range(src_x, dst_x):
            y = src_y if x < med_x else dst_y
            if self.get(x, y).character in ['|', '+']:
                self.write(x, y, '+', color=color, attrs=attrs)
            else:
                self.write(x, y, '-', color=color, attrs=attrs)

        for y in range(min(src_y, dst_y), max(src_y, dst_y)):
            if self.get(med_x, y).character != '+':
                self.write(med_x, y, '|', color=color, attrs=attrs)

        if src_y != dst_y:
            self.write(med_x, src_y, '+', color=color, attrs=attrs)
            self.write(med_x, dst_y, '+', color=color, attrs=attrs)

    def _expand(self, x, y):
        while len(self._cells) < y:
            self._cells.append([])

        while len(self._cells[y - 1]) < x:
            self._cells[y - 1].append(self._default)

    def get(self, x: int, y: int) -> RasterCell:
        if y >= 0 and y < len(self._cells) and x >= 0 and x < len(self._cells[y]):
            return self._cells[y][x]
        else:
            return self._default

    def size(self) -> Tuple[int]:
        if len(self._cells) == 0:
            return 0, 0
        else:
            return max([len(l) for l in self._cells]), len(self._cells)

    def origin_bounds(self, origin: object) -> Boundary:
        raster_w, raster_h = self.size()

        b_x, b_y, b_w, b_h = 0, 0, 0, 0

        for y in range(raster_h):
            for x in range(raster_w):
                cell = self.get(x, y)
                if origin == cell.origin:
                    if b_w == 0:
                        b_x, b_y, b_w, b_h = x, y, 1, 1
                    else:
                        if x >= b_x + b_w:
                            b_w = x - b_x + 1

                        if y >= b_y + b_h:
                            b_h = y - b_y + 1

        return Boundary(b_x, b_y, b_w, b_h) if b_w > 0 else None

    def text(self, color: bool = False):
        text = ''

        for line in self._cells:
            for c in line:
                if color and c.color:
                    text += termcolor.colored(c.character, c.color, attrs=c.attrs)
                else:
                    text += c.character

            text += '\n'

        return text

    def __str__(self):
        return self.text()



#!/usr/bin/env python3
from setuptools import setup
from docker_ascii_map import __version__

setup(
    name='docker-ascii-map',
    version=__version__,
    packages=['docker_ascii_map'],
    package_dir={'docker_ascii_map': 'docker_ascii_map'},
    scripts=['docker_ascii_map/docker-ascii-map'],
    test_suite='tests',
    setup_requires=['pytest-runner'],
    install_requires=['docker-py >= 1.10.0', 'termcolor >= 1.1.0'],
    tests_require=['pytest'],
    url='https://gitlab.com/alcibiade/docker-ascii-map',
    license='MIT',
    author='Yannick Kirschhoffer',
    author_email='alcibiade@alcibiade.org',
    description='A set of python scripts displaying the local docker containers structure and status on an ascii map.',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Topic :: System :: Systems Administration',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
    ],
)



import math

from docker_ascii_map.docker_config import Configuration, Container
from docker_ascii_map.widget import *


def build_container_widget(container: Container, encoding: str) -> Widget:
    lines = []

    if encoding == 'UTF-8':
        statuschar = u"\u2713" if container.is_running() else u'\u2717'
    else:
        statuschar = 'V' if container.is_running() else 'x'

    color = 'green' if container.is_running() else 'red'

    lines.append('[' + statuschar + '] ' + container.name)
    lines.append('    ' + container.image)
    container_widget = Paragraph(lines, color=color)
    return container_widget


def build_ordered_network_list(config: Configuration) -> List[str]:
    networks = set()

    for container in config.containers:
        for net in container.networks:
            networks.add(net)

    networks = list(networks)
    networks.sort()

    # Networks are sorted, now group the ones linked by containers

    for container in config.containers:
        if len(container.networks) > 1:
            base_index = networks.index(container.networks[0])
            for net in container.networks[1:]:
                networks.remove(net)
                base_index += 1
                networks.insert(base_index, net)

    return networks


def build_container_wrapper(container: Container, container_widget: Widget) -> Widget:
    ports_count = len(container.ports)
    total_padding = ports_count - 2

    if total_padding <= 0:
        return container_widget
    else:
        return Padding(container_widget,
                       Size(0, int(math.floor(total_padding / 2))),
                       Size(0, int(math.ceil(total_padding / 2))))


class Renderer:
    def __init__(self):
        pass

    def render_to_raster(self, config: Configuration, encoding: str = 'UTF-8'):
        network_widgets = []

        networks = build_ordered_network_list(config)

        net_widgets_map = {}
        cnt_widgets_map = {}

        # Network boxes with single-network containers

        for net in networks:
            net_widgets = []

            for container in config.containers:
                if [net] == container.networks:
                    container_widget = build_container_widget(container, encoding)
                    cnt_widgets_map[container] = container_widget
                    net_widgets.append(build_container_wrapper(container, container_widget))

            net_box = Border(VBox(net_widgets), net)
            net_widgets_map[net] = net_box
            network_widgets.append(net_box)

        # Containers connected to multiple networks

        bridge_widgets = []
        links = []

        for container in config.containers:
            if len(container.networks) > 1:
                c = Padding(build_container_widget(container, encoding), Size(1, 0))
                cnt_widgets_map[container] = c
                padded = Padding(c, Size(12, 2))
                bridge_widgets.append(padded)

                for n in container.networks:
                    net_box = net_widgets_map[n]
                    links.append((c, net_box))

        networks_box = VBox(network_widgets)
        bridges_box = VBox(bridge_widgets)
        links_box = Links(bridges_box, networks_box, links, True)

        # Port mapping

        portmaps = []

        for container in config.containers:
            for port in container.ports:
                w = cnt_widgets_map[container]
                portmaps.append((w,
                                 str(port.public_port),
                                 'green' if container.is_running() else 'red',
                                 []
                                 ))

        ports_box = Annotations(links_box, portmaps)

        root_box = ports_box
        return root_box.render()

    def render_to_string(self, config: Configuration, encoding: str = 'UTF-8', color: bool = False):
        raster = self.render_to_raster(config, encoding)
        return raster.text(color)

    def render(self, config: Configuration, encoding: str = 'UTF-8', color: bool = False):
        return self.render_to_string(config, encoding, color)
